import re
from threading import RLock
from typing import Any, Dict, Tuple
from urllib.parse import urlparse

from triad.utils.hash import to_uuid

import fs
from fs import memoryfs, open_fs, tempfs
from fs.base import FS as FSBase
from fs.glob import BoundGlobber, Globber
from fs.mountfs import MountFS
from fs.subfs import SubFS

_SCHEME_PREFIX = re.compile(r"^[a-zA-Z0-9\-_]+:")


class FileSystem(MountFS):
    """A unified filesystem based on PyFileSystem2. The special requirement
    for this class is that all paths must be absolute path with scheme.
    To customize different file systems, you should override `create_fs`
    to provide your own configured file systems.

    .. admonition:: Examples

        .. code-block:: python

            fs = FileSystem()
            fs.writetext("mem://from/a.txt", "hello")
            fs.copy("mem://from/a.txt", "mem://to/a.txt")

    .. note::

        If a path is not a local path, it must include the scheme and `netloc`
        (the first element after `://`)
        :param auto_close: If `True` (the default), the child filesystems
        will be closed when `MountFS` is closed.
    """

    def __init__(self, auto_close: bool = True):
        super().__init__(auto_close)
        self._fs_store: Dict[str, FSBase] = {}
        self._in_create = False
        self._fs_lock = RLock()

    def create_fs(self, root: str) -> FSBase:
        """create a PyFileSystem instance from `root`. `root` is in the
        format of `/` if local path, else `<scheme>://<netloc>`.
        You should override this method to provide custom instances, for
        example, if you want to create an S3FS with certain parameters.
        :param root: `/` if local path, else `<scheme>://<netloc>`
        """
        if root.startswith("temp://"):
            fs = tempfs.TempFS(root[len("temp://") :])
            return fs
        if root.startswith("mem://"):
            fs = memoryfs.MemoryFS()
            return fs
        return open_fs(root)

    @property
    def glob(self):
        """A globber object"""
        return _BoundGlobber(self)

    def _delegate(self, path) -> Tuple[FSBase, str]:
        with self._fs_lock:
            if self._in_create:  # pragma: no cover
                return super()._delegate(path)
            self._in_create = True
            fp = _FSPath(path)
            if fp.root not in self._fs_store:
                self._fs_store[fp.root] = self.create_fs(fp.root)
                self.mount(to_uuid(fp.root), self._fs_store[fp.root])
            self._in_create = False
        m_path = to_uuid(fp.root) + "/" + fp.relative_path
        return super()._delegate(m_path)

    def makedirs(
        self, path: str, permissions: Any = None, recreate: bool = False
    ) -> SubFS:
        """Make a directory, and any missing intermediate directories.

        .. note::

            This overrides the base ``makedirs``

        :param path: path to directory from root.
        :param permissions: initial permissions, or `None` to use defaults.
        :recreate: if `False` (the default), attempting to
          create an existing directory will raise an error. Set
          to `True` to ignore existing directories.
        :return: a sub-directory filesystem.

        :raises fs.errors.DirectoryExists: if the path is already
          a directory, and ``recreate`` is `False`.
        :raises fs.errors.DirectoryExpected: if one of the ancestors
          in the path is not a directory.
        """
        self.check()
        fs, _path = self._delegate(path)
        return fs.makedirs(_path, permissions=permissions, recreate=recreate)


class _BoundGlobber(BoundGlobber):
    def __call__(
        self,
        pattern: Any,
        path: str = "/",
        namespaces: Any = None,
        case_sensitive: bool = True,
        exclude_dirs: Any = None,
    ) -> Globber:
        fp = _FSPath(path)
        _path = fs.path.join(fp._root, fp._path) if fp.is_windows else path
        return super().__call__(
            pattern,
            path=_path,
            namespaces=namespaces,
            case_sensitive=case_sensitive,
            exclude_dirs=exclude_dirs,
        )


class _FSPath(object):
    def __init__(self, path: str):
        if path is None:
            raise ValueError("path can't be None")
        path = _modify_path(path)
        self._is_windows = False
        if _is_windows(path):
            self._scheme = ""
            self._root = path[:3]
            self._path = path[3:]
            self._is_windows = True
        elif path.startswith("/"):
            self._scheme = ""
            self._root = "/"
            self._path = fs.path.abspath(path)
        else:
            uri = urlparse(path)
            if uri.scheme == "" and not path.startswith("/"):
                raise ValueError(
                    f"invalid {path}, must be abs path either local or with scheme"
                )
            self._scheme = uri.scheme
            if uri.netloc == "":
                raise ValueError(f"invalid path {path}")
            self._root = uri.scheme + "://" + uri.netloc
            self._path = uri.path
        self._path = self._path.lstrip("/")
        # if self._path == "":
        #    raise ValueError(f"invalid path {path}")

    @property
    def is_windows(self) -> bool:
        return self._is_windows

    @property
    def scheme(self) -> str:
        return self._scheme

    @property
    def root(self) -> str:
        return self._root

    @property
    def relative_path(self) -> str:
        return self._path


def _modify_path(path: str) -> str:  # noqa: C901
    """to fix things like /s3:/a/b.txt -> s3://a/b.txt"""
    if path.startswith("/"):
        s = _SCHEME_PREFIX.search(path[1:])
        if s is not None:
            colon = s.end()
            scheme = path[1:colon]
            if colon + 1 == len(path):  # /C: or /s3:
                path = scheme + "://"
            elif path[colon + 1] == "/":  # /s3:/a/b.txt
                path = scheme + "://" + path[colon + 1 :].lstrip("/")
            elif path[colon + 1] == "\\":  # /c:\a\b.txt
                path = scheme + ":\\" + path[colon + 1 :].lstrip("\\")
    if path.startswith("file:///"):
        path = path[8:]
    elif path.startswith("file://"):
        path = path[6:]
    if path.startswith("\\\\"):
        # windows \\10.100.168.1\...
        raise NotImplementedError(f"path {path} is not supported")
    if path != "" and path[0].isalpha():
        if len(path) == 2 and path[1] == ":":
            # C: => C:/
            return path[0] + ":/"
        if path[1:].startswith(":\\"):
            # C:\a\b\c => C:/a/b/c
            return path[0] + ":/" + path[3:].replace("\\", "/").lstrip("/")
        if path[1:].startswith(":/"):
            # C:/a/b/c => C:/a/b/c
            return path[0] + ":/" + path[3:].lstrip("/")
    return path


def _is_windows(path: str) -> bool:
    if len(path) < 3:
        return False
    return path[0].isalpha() and path[1] == ":" and path[2] == "/"



import itertools
from functools import update_wrapper
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

from .entry_points import load_entry_point


def run_at_def(run_at_def_func: Optional[Callable] = None, **kwargs: Any) -> Callable:
    """Decorator to run the function at declaration. This is useful when we want import
    to trigger a function run (which can guarantee it runs only once).

    .. admonition:: Examples

        Assume the following python file is a module in your package,
        then when you ``import package.module``, the two functions will run.

        .. code-block:: python

            from triad import run_at_def

            @run_at_def
            def register_something():
                print("registered")

            @run_at_def(a=1)
            def register_something2(a):
                print("registered", a)

    :param run_at_def_func: the function to decorate
    :param kwargs: the parameters to call this function
    """

    def _run(_func: Callable) -> Callable:
        _func(**kwargs)
        return _func

    return _run if run_at_def_func is None else _run(run_at_def_func)  # type:ignore


def conditional_dispatcher(
    default_func: Optional[Callable[..., Any]] = None, entry_point: Optional[str] = None
) -> Callable:
    """Decorating a conditional dispatcher that will run the **first matching** registered
    functions in other modules/packages. This is a more general solution compared to
    ``functools.singledispatch``. You can write arbitrary matching functions according
    to all the inputs of the function.

    .. admonition:: Examples

        Assume in ``pkg1.module1``, you have:

        .. code-block:: python

            from triad import conditional_dispatcher

            @conditional_dispatcher(entry_point="my.plugins")
            def get_len(obj):
                raise NotImplementedError

        In another package ``pkg2``, in ``setup.py``, you define
        an entry point as:

        .. code-block:: python

            setup(
                ...,
                entry_points={
                    "my.plugins": [
                        "my = pkg2.module2"
                    ]
                },
            )

        And in ``pkg2.module2``:

        .. code-block:: python

            from pkg1.module1 import get_len

            @get_len.candidate(lambda obj: isinstance(obj, str))
            def get_str_len(obj:str) -> int:
                return len(obj)

            @get_len.candidate(lambda obj: isinstance(obj, int) and obj == 10)
            def get_int_len(obj:int) -> int:
                return obj

        Now, both functions will be automatically registered when ``pkg2``
        is installed in the environement. In another ``pkg3``:

        .. code-block:: python

            from pkg1.module1 import get_len

            assert get_len("abc") == 3  # calling get_str_len
            assert get_len(10) == 10  # calling get_int_len
            get_len(20)  # raise NotImplementedError due to no matching candidates

    .. seealso::

        Please read :meth:`~.ConditionalDispatcher.candidate` for details about the
        matching function and priority settings.

    :param default_func: the function to decorate
    :param entry_point: the entry point to preload dispatchers, defaults to None
    """

    def _run(_func: Callable) -> "ConditionalDispatcher":
        class _Dispatcher(ConditionalDispatcher):
            def __call__(self, *args: Any, **kwds: Any) -> Any:
                return self.run_top(*args, **kwds)

        return _Dispatcher(_func, entry_point=entry_point)

    return _run if default_func is None else _run(default_func)  # type:ignore


def conditional_broadcaster(
    default_func: Optional[Callable[..., Any]] = None, entry_point: Optional[str] = None
) -> Callable:
    """Decorating a conditional broadcaster that will run **all** registered functions in
    other modules/packages.

    .. admonition:: Examples

        Assume in ``pkg1.module1``, you have:

        .. code-block:: python

            from triad import conditional_broadcaster

            @conditional_broadcaster(entry_point="my.plugins")
            def myprint(obj):
                raise NotImplementedError

            @conditional_broadcaster(entry_point="my.plugins")
            def myprint2(obj):
                raise NotImplementedError

        In another package ``pkg2``, in ``setup.py``, you define
        an entry point as:

        .. code-block:: python

            setup(
                ...,
                entry_points={
                    "my.plugins": [
                        "my = pkg2.module2"
                    ]
                },
            )

        And in ``pkg2.module2``:

        .. code-block:: python

            from pkg1.module1 import get_len

            @myprint.candidate(lambda obj: isinstance(obj, str))
            def myprinta(obj:str) -> None:
                print(obj, "a")

            @myprint.candidate(lambda obj: isinstance(obj, str) and obj == "x")
            def myprintb(obj:str) -> None:
                print(obj, "b")

        Now, both functions will be automatically registered when ``pkg2``
        is installed in the environement. In another ``pkg3``:

        .. code-block:: python

            from pkg1.module1 import get_len

            myprint("x")  # calling both myprinta and myprinta
            myprint("y")  # calling myprinta only
            myprint2("x")  # raise NotImplementedError due to no matching candidates

    .. note::

        Only when no matching candidate found, the implementation of the original
        function will be used. If you don't want to throw an error, then use ``pass`` in
        the original function instead.

    .. seealso::

        Please read :meth:`~.ConditionalDispatcher.candidate` for details about the
        matching function and priority settings.

    :param default_func: the function to decorate
    :param entry_point: the entry point to preload dispatchers, defaults to None
    """

    def _run(_func: Callable) -> "ConditionalDispatcher":
        class _Dispatcher(ConditionalDispatcher):
            def __call__(self, *args: Any, **kwds: Any) -> None:
                list(self.run(*args, **kwds))

        return _Dispatcher(_func, entry_point=entry_point)

    return _run if default_func is None else _run(default_func)  # type:ignore


class ConditionalDispatcher:
    """A conditional function dispatcher based on custom matching functions.
    This is a more general solution compared to
    ``functools.singledispatch``. You can write arbitrary matching functions according
    to all the inputs of the function.

    .. note::

        Please use the decorators :func:`.conditional_dispatcher` and
        :func:`.conditional_broadcaster` instead of directly using this class.

    :param default_func: the parent function that will dispatch the execution
        based on matching functions
    :param entry_point: the entry point to preload children functions,
        defaults to None
    """

    def __init__(
        self, default_func: Callable[..., Any], entry_point: Optional[str] = None
    ):
        self._func = default_func
        self._funcs: List[
            Tuple[float, int, Callable[..., bool], Callable[..., Any]]
        ] = []
        self._entry_point = entry_point
        update_wrapper(self, default_func)

    def __getstate__(self) -> Dict[str, Any]:
        return {
            k: v
            for k, v in self.__dict__.items()
            if k in ["_func", "_funcs", "_entry_point"]
        }

    def __setstate__(self, data: Dict[str, Any]) -> None:
        for k, v in data.items():
            setattr(self, k, v)

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        """The abstract method to mimic the function call"""
        raise NotImplementedError  # pragma: no cover

    def run(self, *args: Any, **kwargs: Any) -> Iterable[Any]:
        """Execute all matching children functions as a generator.

        .. note::

            Only when there is matching functions, the default implementation
            will be invoked.
        """
        self._preload()
        has_return = False
        for f in self._funcs:
            if self._match(f[2], *args, **kwargs):
                yield f[3](*args, **kwargs)
                has_return = True
        if not has_return:
            yield self._func(*args, **kwargs)

    def run_top(self, *args: Any, **kwargs: Any) -> Any:
        """Execute the first matching child function

        :return: the return of the child function
        """
        return list(itertools.islice(self.run(*args, **kwargs), 1))[0]

    def register(
        self,
        func: Callable[..., Any],
        matcher: Callable[..., bool],
        priority: float = 1.0,
    ) -> None:
        """Register a child function with matcher and priority.

        .. note::

            The order to be matched is determined by both the priority
            and the order of registration.

            * The default priority is 1.0
            * Children with higher priority values will be matched earlier
            * When ``priority>0`` then later registrations will be matched earlier
            * When ``priority<=0`` then earlier registrations will be matched earlier

            So if you want to 'overwrite' the existed matches, set priority to be
            greater than 1.0. If you want to 'ignore' the current if there are other
            matches, set priority to 0.0.

        :param func: a child function to be used when matching
        :param matcher: a function determines whether it is a match
            based on the same input as the parent function
        :param priority: it determines the order to be matched,
            **higher value means higher priority**, defaults to 1.0
        """
        n = len(self._funcs)
        self._funcs.append((-priority, n if priority <= 0.0 else -n, matcher, func))
        self._funcs.sort()

    def candidate(
        self,
        matcher: Callable[..., bool],
        priority: float = 1.0,
    ) -> Callable:
        """A decorator to register a child function with matcher and priority.

        .. note::

            The order to be matched is determined by both the priority
            and the order of registration.

            * The default priority is 1.0
            * Children with higher priority values will be matched earlier
            * When ``priority>0`` then later registrations will be matched earlier
            * When ``priority<=0`` then earlier registrations will be matched earlier

            So if you want to 'overwrite' the existed matches, set priority to be
            greater than 1.0. If you want to 'ignore' the current if there are other
            matches, set priority to 0.0.

        .. seealso::

            Please see examples in :func:`.conditional_dispatcher` and
            :func:`.conditional_broadcaster`.

        :param matcher: a function determines whether it is a match
            based on the same input as the parent function
        :param priority: it determines the order to be matched,
            **higher value means higher priority**, defaults to 1.0
        """

        def _run(_func: Callable[..., Any]):
            self.register(_func, matcher=matcher, priority=priority)
            return _func

        return _run

    def _preload(self) -> None:
        if self._entry_point is not None:
            load_entry_point(self._entry_point)

    def _match(self, m: Callable[..., bool], *args: Any, **kwargs: Any) -> bool:
        try:
            return m(*args, **kwargs)
        except Exception:
            return False



from triad.utils.convert import to_type


class BaseClass(object):
    pass


class SubClass(BaseClass):
    pass


class Class2(object):
    def __init__(self, a=0, b=0, c=0):
        self.s = str(a) + str(b) + str(c)


class __Dummy__(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def invoke_to_type(exp):
    return to_type(exp)



import json
from typing import Any, Dict, Hashable, List, Tuple


def loads_no_dup(json_str: str) -> Any:
    """Load json string, and raise KeyError if there are duplicated keys

    :param json_str: json string
    :raises KeyError: if there are duplicated keys
    :return: the parsed object
    """
    return json.loads(json_str, object_pairs_hook=check_for_duplicate_keys)


def check_for_duplicate_keys(
    ordered_pairs: List[Tuple[Hashable, Any]]
) -> Dict[Any, Any]:
    """Raise ValueError if a duplicate key exists in provided ordered list of pairs,
    otherwise return a dict.

    Example:

    >>> json.loads('{"x": 1, "x": 2}', object_pairs_hook=check_for_duplicate_keys)

    :raises KeyError: if there is duplicated key
    """
    dict_out: Dict[Any, Any] = {}
    for key, val in ordered_pairs:
        if key in dict_out:
            raise KeyError(f"Duplicate key: {key}")
        else:
            dict_out[key] = val
    return dict_out



import copy
import json
import sys
from collections import OrderedDict
from typing import Any, Dict, List, Tuple, TypeVar, Union

from triad.exceptions import InvalidOperationError
from triad.utils.assertion import assert_arg_not_none
from triad.utils.convert import as_type
from triad.utils.iter import to_kv_iterable

KT = TypeVar("KT")
VT = TypeVar("VT")


class IndexedOrderedDict(OrderedDict, Dict[KT, VT]):
    """Subclass of OrderedDict that can get and set with index"""

    def __init__(self, *args: Any, **kwds: Any):
        self._readonly = False
        self._need_reindex = True
        self._key_index: Dict[Any, int] = {}
        self._index_key: List[Any] = []
        super().__init__(*args, **kwds)

    @property
    def readonly(self) -> bool:
        """Whether this dict is readonly"""
        return self._readonly

    def set_readonly(self) -> None:
        """Make this dict readonly"""
        self._readonly = True

    def index_of_key(self, key: Any) -> int:
        """Get index of key

        :param key: key value
        :return: index of the key value
        """
        self._build_index()
        return self._key_index[key]

    def get_key_by_index(self, index: int) -> KT:
        """Get key by index

        :param index: index of the key
        :return: key value at the index
        """
        self._build_index()
        return self._index_key[index]

    def get_value_by_index(self, index: int) -> VT:
        """Get value by index

        :param index: index of the item
        :return: value at the index
        """
        key = self.get_key_by_index(index)
        return self[key]

    def get_item_by_index(self, index: int) -> Tuple[KT, VT]:
        """Get key value pair by index

        :param index: index of the item
        :return: key value tuple at the index
        """
        key = self.get_key_by_index(index)
        return key, self[key]

    def set_value_by_index(self, index: int, value: VT) -> None:
        """Set value by index

        :param index: index of the item
        :param value: new value
        """
        key = self.get_key_by_index(index)
        self[key] = value

    def pop_by_index(self, index: int) -> Tuple[KT, VT]:
        """Pop item at index

        :param index: index of the item
        :return: key value tuple at the index
        """
        key = self.get_key_by_index(index)
        return key, self.pop(key)

    def equals(self, other: Any, with_order: bool):
        """Compare with another object

        :param other: for possible types, see :func:`~triad.utils.iter.to_kv_iterable`
        :param with_order: whether to compare order
        :return: whether they equal
        """
        if with_order:
            if isinstance(other, OrderedDict):
                return self == other
            return self == OrderedDict(to_kv_iterable(other))
        else:
            if isinstance(other, OrderedDict) or not isinstance(other, Dict):
                return self == dict(to_kv_iterable(other))
            return self == other

    # ----------------------------------- Wrappers over OrderedDict

    def __setitem__(  # type: ignore
        self, key: KT, value: VT, *args: Any, **kwds: Any
    ) -> None:
        self._pre_update("__setitem__", key not in self)
        super().__setitem__(key, value, *args, **kwds)  # type: ignore

    def __delitem__(self, *args: Any, **kwds: Any) -> None:  # type: ignore
        self._pre_update("__delitem__")
        super().__delitem__(*args, **kwds)  # type: ignore

    def clear(self) -> None:
        self._pre_update("clear")
        super().clear()

    def copy(self) -> "IndexedOrderedDict":
        other = super().copy()
        assert isinstance(other, IndexedOrderedDict)
        other._need_reindex = self._need_reindex
        other._index_key = self._index_key.copy()
        other._key_index = self._key_index.copy()
        other._readonly = False
        return other

    def __copy__(self) -> "IndexedOrderedDict":
        return self.copy()

    def __deepcopy__(self, arg: Any) -> "IndexedOrderedDict":
        it = [(copy.deepcopy(k), copy.deepcopy(v)) for k, v in self.items()]
        return IndexedOrderedDict(it)

    def popitem(self, *args: Any, **kwds: Any) -> Tuple[KT, VT]:  # type: ignore
        self._pre_update("popitem")
        return super().popitem(*args, **kwds)  # type: ignore

    def move_to_end(self, *args: Any, **kwds: Any) -> None:  # type: ignore
        self._pre_update("move_to_end")
        super().move_to_end(*args, **kwds)  # type: ignore

    def __sizeof__(self) -> int:  # pragma: no cover
        return super().__sizeof__() + sys.getsizeof(self._need_reindex)

    def pop(self, *args: Any, **kwds: Any) -> VT:  # type: ignore
        self._pre_update("pop")
        return super().pop(*args, **kwds)  # type: ignore

    def _build_index(self) -> None:
        if self._need_reindex:
            self._index_key = list(self.keys())
            self._key_index = {x: i for i, x in enumerate(self._index_key)}
            self._need_reindex = False

    def _pre_update(self, op: str, need_reindex: bool = True) -> None:
        if self.readonly:
            raise InvalidOperationError("This dict is readonly")
        self._need_reindex = need_reindex


class ParamDict(IndexedOrderedDict[str, Any]):
    """Parameter dictionary, a subclass of ``IndexedOrderedDict``, keys must be string

    :param data: for possible types, see :func:`~triad.utils.iter.to_kv_iterable`
    :param deep: whether to deep copy ``data``
    """

    OVERWRITE = 0
    THROW = 1
    IGNORE = 2

    def __init__(self, data: Any = None, deep: bool = True):
        super().__init__()
        self.update(data, deep=deep)

    def __setitem__(  # type: ignore
        self, key: str, value: Any, *args: Any, **kwds: Any
    ) -> None:
        assert isinstance(key, str)
        super().__setitem__(key, value, *args, **kwds)  # type: ignore

    def __getitem__(self, key: Union[str, int]) -> Any:  # type: ignore
        if isinstance(key, int):
            key = self.get_key_by_index(key)
        return super().__getitem__(key)  # type: ignore

    def get(self, key: Union[int, str], default: Any) -> Any:  # type: ignore
        """Get value by ``key``, and the value must be a subtype of the type of
        ``default``(which can't be None). If the ``key`` is not found,
        return ``default``.

        :param key: the key to search
        :raises NoneArgumentError: if default is None
        :raises TypeError: if the value can't be converted to the type of ``default``

        :return: the value by ``key``, and the value must be a subtype of the type of
            ``default``. If ``key`` is not found, return `default`
        """
        assert_arg_not_none(default, "default")
        if (isinstance(key, str) and key in self) or isinstance(key, int):
            return as_type(self[key], type(default))
        return default

    def get_or_none(self, key: Union[int, str], expected_type: type) -> Any:
        """Get value by `key`, and the value must be a subtype of ``expected_type``

        :param key: the key to search
        :param expected_type: expected return value type

        :raises TypeError: if the value can't be converted to ``expected_type``

        :return: if ``key`` is not found, None. Otherwise if the value can be converted
            to ``expected_type``, return the converted value, otherwise raise exception
        """
        return self._get_or(key, expected_type, throw=False)

    def get_or_throw(self, key: Union[int, str], expected_type: type) -> Any:
        """Get value by ``key``, and the value must be a subtype of ``expected_type``.
        If ``key`` is not found or value can't be converted to ``expected_type``, raise
        exception

        :param key: the key to search
        :param expected_type: expected return value type

        :raises KeyError: if ``key`` is not found
        :raises TypeError: if the value can't be converted to ``expected_type``

        :return: only when ``key`` is found and can be converted to ``expected_type``,
            return the converted value
        """
        return self._get_or(key, expected_type, throw=True)

    def to_json(self, indent: bool = False) -> str:
        """Generate json expression string for the dictionary

        :param indent: whether to have indent
        :return: json string
        """
        if not indent:
            return json.dumps(self, separators=(",", ":"))
        else:
            return json.dumps(self, indent=4)

    def update(  # type: ignore
        self, other: Any, on_dup: int = 0, deep: bool = True
    ) -> "ParamDict":
        """Update dictionary with another object (for possible types,
        see :func:`~triad.utils.iter.to_kv_iterable`)

        :param other: for possible types, see :func:`~triad.utils.iter.to_kv_iterable`
        :param on_dup: one of ``ParamDict.OVERWRITE``, ``ParamDict.THROW``
            and ``ParamDict.IGNORE``

        :raises KeyError: if using ``ParamDict.THROW`` and other contains existing keys
        :raises ValueError: if ``on_dup`` is invalid
        :return: itself
        """
        self._pre_update("update", True)
        for k, v in to_kv_iterable(other):
            if on_dup == ParamDict.OVERWRITE or k not in self:
                self[k] = copy.deepcopy(v) if deep else v
            elif on_dup == ParamDict.THROW:
                raise KeyError(f"{k} exists in dict")
            elif on_dup == ParamDict.IGNORE:
                continue
            else:
                raise ValueError(f"{on_dup} is not supported")
        return self

    def _get_or(
        self, key: Union[int, str], expected_type: type, throw: bool = True
    ) -> Any:
        if (isinstance(key, str) and key in self) or isinstance(key, int):
            return as_type(self[key], expected_type)
        if throw:
            raise KeyError(f"{key} not found")
        return None



from functools import _make_key, wraps
from threading import RLock
from typing import Any, Callable, Dict, Optional, Tuple, Type


class SerializableRLock:
    """A serialization safe wrapper of :external+python:class:`threading.RLock`"""

    def __init__(self):
        self._lock = RLock()

    def __enter__(self) -> Any:
        return self._lock.__enter__()

    def __exit__(
        self, exception_type: Any, exception_value: Any, exception_traceback: Any
    ) -> Any:
        return self._lock.__exit__(exception_type, exception_value, exception_traceback)

    def __getstate__(self) -> Dict[str, Any]:
        return {}

    def __setstate__(self, data: Dict[str, Any]) -> None:
        self._lock = RLock()


def run_once(
    func: Optional[Callable] = None,
    key_func: Optional[Callable] = None,
    lock_type: Type = RLock,
) -> Callable:
    """The decorator to run `func` once, the uniqueness is defined by `key_func`.
    This implementation is serialization safe and thread safe.

    :param func: the function to run only once with this wrapper instance
    :param key_func: the unique key determined by arguments of `func`, if not set, it
      will use the same hasing logic as :external+python:func:`functools.lru_cache`
    :param lock_type: lock class type for thread safe, it doesn't need to be
      serialization safe

    .. admonition:: Examples

        .. code-block:: python

            @run_once
            def r(a):
                return max(a)

            a1 = [0, 1]
            a2 = [0, 2]
            assert 1 == r(a1) # will trigger r
            assert 1 == r(a1) # will get the result from cache
            assert 2 == r(a2) # will trigger r again because of different arguments

            # the following example ignores arguments
            @run_once(key_func=lambda *args, **kwargs: True)
            def r2(a):
                return max(a)

            assert 1 == r(a1) # will trigger r
            assert 1 == r(a2) # will get the result from cache

    .. note::

        * Hash collision is the concern of the user, not this class, your
          `key_func` should avoid any potential collision
        * `func` can have no return
        * For concurrent calls of this wrapper, only one will trigger `func` other
          calls will be blocked until the first call returns an result
        * This class is cloudpicklable, but unpickled instance does NOT share the same
          context with the original one
        * This is not to replace :external+python:func:`functools.lru_cache`,
          it is not supposed to cache a lot of items
    """

    def _run(func: Callable) -> "RunOnce":
        return wraps(func)(RunOnce(func, key_func=key_func, lock_type=lock_type))

    return _run(func) if func is not None else wraps(func)(_run)  # type: ignore


class RunOnce:
    """Run `func` once, the uniqueness is defined by `key_func`.
    This implementation is serialization safe and thread safe.

    .. note::

        Please use the decorator :func:`~.run_once` instead of directly
        using this class

    :param func: the function to run only once with this wrapper instance
    :param key_func: the unique key determined by arguments of `func`, if not set, it
        will use the same hasing logic as :external+python:func:`functools.lru_cache`
    :param lock_type: lock class type for thread safe
    """

    def __init__(
        self,
        func: Callable,
        key_func: Optional[Callable] = None,
        lock_type: Type = RLock,
    ):
        self._func = func
        if key_func is None:
            self._key_func: Callable = lambda *args, **kwargs: _make_key(
                args, kwargs, typed=True
            )
        else:
            self._key_func = key_func
        self._lock_type = lock_type
        self._init_locks()

    def __getstate__(self) -> Dict[str, Any]:
        d = dict(self.__dict__)
        del d["_lock"]
        del d["_locks"]
        del d["_store"]
        return d

    def __setstate__(self, members: Any) -> None:
        self.__dict__.update(members)
        self._init_locks()

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        key = self._key_func(*args, **kwargs)
        lock = self._get_lock(key)
        with lock:
            found, res = self._try_get(key)
            if found:
                return res
            res = self._func(*args, **kwargs)
            self._update(key, res)
            return res

    def _get_lock(self, key) -> Any:
        with self._lock:
            if key not in self._locks:
                self._locks[key] = self._lock_type()
            return self._locks[key]

    def _try_get(self, key) -> Tuple[bool, Any]:
        with self._lock:
            if key in self._store:
                return True, self._store[key]
            return False, None

    def _update(self, key, value) -> None:
        with self._lock:
            self._store[key] = value

    def _init_locks(self):
        self._lock = self._lock_type()
        self._locks: Dict[int, Any] = {}
        self._store: Dict[int, Any] = {}
